@@ -342,7 +342,11 @@ static size_t dopr(char *buffer, size_t maxlen, const char *format, va_list args
 				break;
 			case 'p':
 				strvalue = (char *)va_arg(args, void *);
+#ifdef _WIN32
+				fmtint (buffer, &currlen, maxlen, (intptr_t) strvalue, 16, min, max, flags);
+#else
 				fmtint (buffer, &currlen, maxlen, (long) strvalue, 16, min, max, flags);
+#endif
 				break;
 			case 'n':
 				if (cflags == DP_C_SHORT) {
